from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from amrita_core.tools.mcp import (
    NOT_GIVEN,
    ClientManager,
    MCPClient,
    MultiClientManager,
)


class TestMCPClient:
    @pytest.fixture
    def mcp_client(self):
        return MCPClient(server_script="test_script")

    def test_initialization(self, mcp_client: MCPClient):
        assert mcp_client.mcp_client is None
        assert mcp_client.server_script == "test_script"
        assert mcp_client.tools == []
        assert mcp_client.openai_tools == []

    @pytest.mark.asyncio
    async def test_aenter_and_aexit(self, mcp_client: MCPClient):
        with patch.object(mcp_client, "_connect") as mock_connect:
            result = await mcp_client.__aenter__()
            assert result == mcp_client
            mock_connect.assert_called_once()

    @pytest.mark.asyncio
    @patch("amrita_core.tools.mcp.Client")
    async def test_connect(self, mock_client_class):
        # Create a mock client instance
        mock_client_instance = AsyncMock()
        mock_client_class.return_value = mock_client_instance

        client = MCPClient(server_script="test_script")

        # Mock the list_tools return value
        mock_client_instance.list_tools.return_value = []

        # Connect manually
        await client._connect(update_tools=True)

        # Verify that the connection was successful
        assert client.mcp_client is not None

    @pytest.mark.asyncio
    async def test_simple_call(self, mcp_client):
        # Test simple call
        with (
            patch.object(mcp_client, "_connect"),
            patch.object(mcp_client, "_close") as mock_close,
        ):
            # Mock mcp_client instance
            mock_client = AsyncMock()
            mock_result = MagicMock()
            mock_result.content = []
            mock_client.call_tool.return_value = mock_result

            mcp_client.mcp_client = mock_client

            await mcp_client.simple_call("test_tool", {"param": "value"})

            # Verify that the tool was called correctly
            mock_client.call_tool.assert_called_once_with(
                "test_tool", {"param": "value"}
            )
            mock_close.assert_called_once()


class TestMultiClientManager:
    @pytest.fixture
    def manager(self):
        # Create a new instance to avoid state pollution
        manager = MultiClientManager()
        manager.clients = []
        manager.script_to_clients = {}
        manager.name_to_clients = {}
        manager.tools_remapping = {}
        manager.reversed_remappings = {}
        return manager

    def test_initialization(self):
        manager = MultiClientManager()
        assert manager.clients == []
        assert manager.script_to_clients == {}
        assert manager.name_to_clients == {}
        assert manager.tools_remapping == {}
        assert manager.reversed_remappings == {}
        assert hasattr(manager, "tools_manager")
        assert manager._is_initialized is False

    def test_get_client_by_script(self):
        manager = MultiClientManager()
        client = manager.get_client_by_script("test_script")
        assert isinstance(client, MCPClient)
        assert client.server_script == "test_script"

    @pytest.mark.asyncio
    async def test_register_only_with_server_script(self):
        manager = MultiClientManager()
        result = manager.register_only(server_script="test_script")
        assert result == manager
        assert len(manager.clients) == 1
        assert isinstance(manager.clients[0], MCPClient)

    @pytest.mark.asyncio
    async def test_register_only_with_client(self):
        manager = MultiClientManager()
        client = MCPClient(server_script="test_script")
        result = manager.register_only(client=client)
        assert result == manager
        assert len(manager.clients) == 1
        assert manager.clients[0] == client

    @pytest.mark.asyncio
    async def test_initialize_this(self):
        manager = MultiClientManager()
        with patch.object(manager, "_load_this") as mock_load:
            result = await manager.initialize_this("test_script")
            assert result == manager
            mock_load.assert_called_once()

    def test_tools_wrapper(self):
        manager = MultiClientManager()
        # Test tools wrapper
        wrapper = manager._tools_wrapper("test_tool")
        assert callable(wrapper)


class TestClientManager:
    def test_singleton_behavior(self):
        # Reset ClientManager state to ensure test accuracy
        ClientManager._instance = None
        ClientManager._initialized = False

        manager1 = ClientManager()
        manager2 = ClientManager()
        assert manager1 is manager2

    def test_inheritance_from_multi_client_manager(self):
        # Reset ClientManager state
        ClientManager._instance = None
        ClientManager._initialized = False

        manager = ClientManager()
        assert isinstance(manager, MultiClientManager)


def test_not_given():
    # Test NOT_GIVEN class
    assert NOT_GIVEN is not None
    assert isinstance(NOT_GIVEN, type)



from unittest.mock import MagicMock

import pytest

from amrita_core.tools.manager import (
    MultiToolsManager,
    ToolsManager,
    _parse_google_docstring,
    _python_type_to_json_type,
    on_tools,
    simple_tool,
)
from amrita_core.tools.models import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    ToolData,
    ToolFunctionSchema,
)


class TestMultiToolsManager:
    @pytest.fixture
    def manager(self) -> MultiToolsManager:
        manager = MultiToolsManager()
        manager._models = {}
        manager._disabled_tools = set()
        return manager

    def test_initialization(self):
        manager = MultiToolsManager()
        assert manager._models == {}
        assert manager._disabled_tools == set()
        assert manager._inited is True

    def test_has_tool(self, manager: MultiToolsManager):
        assert not manager.has_tool("nonexistent_tool")

        tool_data = MagicMock()
        tool_data.data.function.name = "test_tool"
        manager.register_tool(tool_data)

        assert manager.has_tool("test_tool")
        assert not manager.has_tool("other_tool")

    def test_register_and_get_tool(self, manager: MultiToolsManager):
        func_mock = MagicMock()
        func_mock.return_value = "test result"

        function_def = FunctionDefinitionSchema(
            name="test_tool",
            description="Test tool description",
            parameters=FunctionParametersSchema(type="object", properties={}),
        )

        tool_data = ToolData(
            func=func_mock,
            data=ToolFunctionSchema(
                function=function_def, type="function", strict=False
            ),
        )

        manager.register_tool(tool_data)

        retrieved_tool = manager.get_tool("test_tool")
        assert retrieved_tool is not None
        assert retrieved_tool == tool_data
        assert retrieved_tool.func == func_mock

        assert manager.get_tool("nonexistent_tool") is None
        assert manager.get_tool("nonexistent_tool", "default") == "default"

    def test_remove_tool(self, manager: MultiToolsManager):
        tool_data = MagicMock()
        tool_data.data.function.name = "test_tool"
        manager.register_tool(tool_data)

        assert manager.has_tool("test_tool")

        manager.remove_tool("test_tool")

        assert not manager.has_tool("test_tool")

    def test_disable_enable_tool(self, manager: MultiToolsManager):
        tool_data = MagicMock()
        tool_data.data.function.name = "test_tool"
        manager.register_tool(tool_data)

        assert manager.has_tool("test_tool")
        assert manager.get_tool("test_tool") is not None

        manager.disable_tool("test_tool")

        assert not manager.has_tool("test_tool")
        assert manager.get_tool("test_tool") is None

        manager.enable_tool("test_tool")

        assert manager.has_tool("test_tool")
        assert manager.get_tool("test_tool") is not None

    def test_get_disabled_tools(self, manager: MultiToolsManager):
        tool_data = MagicMock()
        tool_data.data.function.name = "test_tool"
        manager.register_tool(tool_data)

        manager.disable_tool("test_tool")

        disabled_tools = manager.get_disabled_tools()
        assert "test_tool" in disabled_tools

    def test_get_tools_and_meta(self, manager: MultiToolsManager):
        tool_data = MagicMock()
        tool_data.data.function.name = "test_tool"
        manager.register_tool(tool_data)

        tools = manager.get_tools()
        assert "test_tool" in tools
        assert tools["test_tool"] == tool_data

        meta = manager.tools_meta()
        assert "test_tool" in meta

        meta_dict = manager.tools_meta_dict()
        assert "test_tool" in meta_dict


class TestToolsManagerSingleton:
    def test_singleton_behavior(self):
        manager1 = ToolsManager()
        manager2 = ToolsManager()
        assert manager1 is manager2

        tool_data = MagicMock()
        tool_data.data.function.name = "shared_tool"
        manager1.register_tool(tool_data)

        assert manager2.has_tool("shared_tool")


def test_parse_google_docstring():
    def sample_function():
        """This is a sample function.

        Args:
            param1 (int): The first parameter.
            param2 (str): The second parameter.

        Returns:
            bool: The return value.
        """
        pass

    desc, params = _parse_google_docstring(sample_function.__doc__)

    assert "This is a sample function" in desc
    assert "param1" in params
    assert "param2" in params
    assert params["param1"] == "The first parameter."
    assert params["param2"] == "The second parameter."


def test_python_type_to_json_type():
    assert _python_type_to_json_type(str) == "string"
    assert _python_type_to_json_type(int) == "integer"
    assert _python_type_to_json_type(float) == "number"
    assert _python_type_to_json_type(bool) == "boolean"
    assert _python_type_to_json_type(list) == "array"
    assert _python_type_to_json_type(dict) == "object"

    assert _python_type_to_json_type(type) == "string"


def test_on_tools_decorator():
    function_def = FunctionDefinitionSchema(
        name="decorated_tool",
        description="A decorated tool",
        parameters=FunctionParametersSchema(type="object", properties={}),
    )

    @on_tools(function_def)
    async def test_function(params):
        return "result"

    manager = ToolsManager()
    tool = manager.get_tool("decorated_tool")
    assert tool is not None
    assert tool.data.function.name == "decorated_tool"

    manager.remove_tool("decorated_tool")


@pytest.mark.asyncio
async def test_simple_tool_decorator():
    @simple_tool
    def add_numbers(a: int, b: int) -> int:
        """Add two numbers together.

        Args:
            a (int): The first number.
            b (int): The second number.

        Returns:
            int: The sum of the two numbers.
        """
        return a + b

    manager = ToolsManager()
    tool = manager.get_tool("add_numbers")

    if tool is not None:
        result = await tool.func({"a": 5, "b": 3})  # type: ignore
        assert result == "8"

    manager.remove_tool("add_numbers")


def test_tool_data_access_methods():
    manager = MultiToolsManager()

    func_mock = MagicMock()
    function_def = FunctionDefinitionSchema(
        name="test_method_tool",
        description="Test tool for method access",
        parameters=FunctionParametersSchema(type="object", properties={}),
    )

    tool_data = ToolData(
        func=func_mock,
        data=ToolFunctionSchema(function=function_def, type="function", strict=False),
        custom_run=False,
    )

    manager.register_tool(tool_data)

    assert manager.get_tool_meta("test_method_tool") is not None
    assert manager.get_tool_func("test_method_tool") is not None

    assert manager.get_tool("nonexistent", "default") == "default"
    assert manager.get_tool_meta("nonexistent", "default") == "default"
    assert manager.get_tool_func("nonexistent", "default") == "default"

    manager.remove_tool("test_method_tool")



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from amrita_core.builtins.agent import (
    AGENT_PROCESS_TOOLS,
    BUILTIN_TOOLS_NAME,
    agent_core,
    cookie,
)
from amrita_core.chatmanager import ChatObject
from amrita_core.config import AmritaConfig, FunctionConfig
from amrita_core.hook.event import CompletionEvent, PreCompletionEvent
from amrita_core.types import Message, UniResponse


@pytest.fixture
def mock_config():
    config = AmritaConfig()
    config.function_config = FunctionConfig()
    return config


@pytest.fixture
def mock_event():
    event: PreCompletionEvent = MagicMock(spec=PreCompletionEvent)
    event.chat_object = MagicMock(spec=ChatObject)
    event.chat_object.session_id = "test-session"
    event.chat_object.preset = "default-preset"
    event.message = MagicMock()
    event.message.train = Message(role="system", content="System message")
    event.message.user_query = Message(role="user", content="User query")
    event.message.unwrap = MagicMock(
        return_value=[
            Message(role="system", content="System message"),
            Message(role="user", content="User query"),
        ]
    )
    event.original_context = "Original context"
    event.get_context_messages = MagicMock()
    event.get_context_messages.return_value = MagicMock()
    event.get_context_messages.return_value.train = MagicMock()
    event.get_context_messages.return_value.train.content = "Context content"
    event._context_messages = []
    return event


@pytest.mark.asyncio
async def test_builtin_tools_constants():
    assert len(BUILTIN_TOOLS_NAME) > 0
    assert isinstance(BUILTIN_TOOLS_NAME, set)

    assert len(AGENT_PROCESS_TOOLS) > 0
    assert isinstance(AGENT_PROCESS_TOOLS, tuple)


@pytest.mark.asyncio
@patch("amrita_core.builtins.agent.SessionsManager")
@patch("amrita_core.builtins.agent.get_config")
@patch("amrita_core.builtins.agent.tools_caller")
async def test_agent_core_basic(
    mock_tools_caller, mock_get_config, mock_sessions_manager
):
    mock_config: AmritaConfig = MagicMock()
    mock_config.function_config.tool_calling_mode = "agent"
    mock_config.function_config.agent_thought_mode = "reasoning"
    mock_config.function_config.agent_tool_call_limit = 5
    mock_config.function_config.agent_tool_call_notice = "notify"
    mock_config.function_config.use_minimal_context = False
    mock_config.function_config.agent_reasoning_hide = False
    mock_config.function_config.agent_middle_message = True
    mock_get_config.return_value = mock_config

    mock_session = MagicMock()
    mock_session.tools = MagicMock()
    mock_session.tools.tools_meta_dict.return_value = {}
    mock_sessions_manager.return_value.get_session_data.return_value = mock_session

    event = MagicMock(spec=PreCompletionEvent)
    event.chat_object = MagicMock(spec=ChatObject)
    event.chat_object.session_id = "test-session"
    event.chat_object.preset = "default-preset"
    event.chat_object.yield_response = AsyncMock()
    event.chat_object.set_queue_done = AsyncMock()
    event.message = MagicMock()
    event.message.train = Message(role="system", content="System message")
    event.message.user_query = Message(role="user", content="User query")
    event.message.unwrap = MagicMock(
        return_value=[
            Message(role="system", content="System message"),
            Message(role="user", content="User query"),
        ]
    )
    event.original_context = "Original context"
    event.get_context_messages = MagicMock()
    event.get_context_messages.return_value = MagicMock()
    event.get_context_messages.return_value.train = MagicMock()
    event.get_context_messages.return_value.train.content = "Context content"
    event._context_messages = []

    mock_response = MagicMock(spec=UniResponse)
    mock_response.content = None
    mock_response.tool_calls = None
    mock_tools_caller.return_value = mock_response

    await agent_core(event, mock_config)

    mock_sessions_manager.assert_called_once()


@pytest.mark.asyncio
@patch("amrita_core.builtins.agent.get_config")
async def test_cookie_handler(mock_get_config):
    mock_config = MagicMock()
    mock_config.cookie.enable_cookie = True
    mock_config.cookie.cookie = "test-cookie"
    mock_get_config.return_value = mock_config

    event = MagicMock(spec=CompletionEvent)
    event.chat_object = MagicMock()
    event.chat_object.yield_response = AsyncMock()
    event.chat_object.set_queue_done = AsyncMock()
    event.get_model_response = MagicMock(
        return_value="Some response with test-cookie inside"
    )

    await cookie(event, mock_config)

    event.chat_object.yield_response.assert_called()
    event.chat_object.set_queue_done.assert_called()


@pytest.mark.asyncio
@patch("amrita_core.builtins.agent.get_config")
async def test_cookie_handler_no_match(mock_get_config):
    mock_config = MagicMock()
    mock_config.cookie.enable_cookie = True
    mock_config.cookie.cookie = "test-cookie"
    mock_get_config.return_value = mock_config

    event = MagicMock(spec=CompletionEvent)
    event.chat_object = MagicMock()
    event.chat_object.yield_response = AsyncMock()
    event.chat_object.set_queue_done = AsyncMock()
    event.get_model_response = MagicMock(return_value="Response without cookie")

    await cookie(event, mock_config)

    event.chat_object.yield_response.assert_not_called()
    event.chat_object.set_queue_done.assert_not_called()
